import logging

from fastapi import (Depends, APIRouter, HTTPException, status, responses,
                     UploadFile)
from fastapi.responses import FileResponse

from gisaf.models.admin import AdminBasket, BasketImportResult, BasketNameOnly
from gisaf.models.authentication import User, UserRead
from gisaf.security import get_current_active_user
from gisaf.admin import manager

logger = logging.getLogger(__name__)

api = APIRouter(
    tags=["admin"],
    # dependencies=[Depends(get_token_header)],
    responses={404: {"description": "Not found"}},
)

@api.get('/basket')
async def get_baskets(
    user: User = Depends(get_current_active_user),
    ) -> list[BasketNameOnly]:
    return [
        BasketNameOnly(name=name)
        for name, basket in (await manager.baskets_for_role(user)).items()
    ]

@api.get('/basket/{name}')
async def get_basket(
    name: str,
    user: User = Depends(get_current_active_user),
    ) -> AdminBasket:
    basket = manager.baskets[name]
    if basket.role and not user.has_role(basket.role):
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    return AdminBasket(
        name=name,
        files=await basket.get_files(),
        columns=basket.columns,
        uploadFields=basket.upload_fields,
        ## TODO: Fix projects
        # projects=getattr(basket, 'projects', None)
    )

@api.post('/basket/upload/{name}')
async def upload_basket_file(
    name: str,
    file: UploadFile,
    project_id: int | None = None,
    surveyor_id: int | None = None,
    equipment_id: int | None = None,
    auto_import: bool = False,
    user: UserRead = Depends(get_current_active_user),
) -> BasketImportResult:
    try:
        basket = manager.baskets[name]
    except KeyError:
        raise HTTPException(status.HTTP_404_NOT_FOUND, f"No basket named {name}")
    fileItem = await basket.add_file(
        file,
        user,
        project_id=project_id,
        surveyor_id=surveyor_id,
        equipment_id=equipment_id,
    )
    return fileItem

@api.get('/basket/download/{name}/{file_id}/{file_name}')
async def download_basket_file(
    name: str,
    file_id: int,
    file_name: str,
    user: User = Depends(get_current_active_user),
    ) -> FileResponse:
    try:
        basket = manager.baskets[name]
    except KeyError:
        raise HTTPException(status.HTTP_404_NOT_FOUND, f'No basket named {name}')
    if basket.role:
        if not user.has_role(basket.role):
            raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    file_record = await basket.get_file(file_id)
    if file_record is None:
        raise HTTPException(status.HTTP_404_NOT_FOUND, f'No import file id {file_id}')
    abs_path = basket.base_dir / file_record.path
    if not abs_path.exists():
        raise HTTPException(status.HTTP_404_NOT_FOUND, f'File {file_record.name} not found')
    return FileResponse(abs_path)

@api.get('/basket/import/{basket}/{file_id}')
async def import_basket_file(
    basket: str,
    file_id: int,
    dryRun: bool = False,
    user: User = Depends(get_current_active_user),
    ) -> BasketImportResult:
    if not (user and user.has_role('reviewer')):
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    basket_ = manager.baskets[basket]
    file_import = await basket_.get_file(file_id)
    if file_import is None:
        raise HTTPException(status.HTTP_404_NOT_FOUND, f'No import file id {file_id}')
    try:
        result = await basket_.import_file(file_import, dryRun)
    ## FIXME: shouldn't it be AdminImportError?
    except ImportError as err:
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, err)
    return result


@api.get('/basket/delete/{basket}/{file_id}')
async def delete_basket_file(
    basket: str,
    file_id: int,
    user: User = Depends(get_current_active_user),
    ) -> None:
    if not (user and user.has_role('reviewer')):
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    basket_ = manager.baskets[basket]
    file_import = await basket_.get_file(file_id)
    if file_import is None:
        raise HTTPException(status.HTTP_404_NOT_FOUND, f'No import file id {file_id}')
    await basket_.delete_file(file_import)