import logging
import io
from typing import Annotated

from fastapi import (Depends, APIRouter, HTTPException, status, Response)
from fastapi.responses import FileResponse, StreamingResponse
from sqlmodel import select
from sqlalchemy.orm import selectinload, joinedload

from gisaf.database import pandas_query, fastapi_db_session as db_session
from gisaf.models.geo_models_base import GeoModel, PlottableModel
from gisaf.models.info import Downloader
from gisaf.security import (
    Token, authenticate_user, get_current_active_user, create_access_token,
    )
from gisaf.models.authentication import (User, UserRead, Role, RoleRead)
from gisaf.registry import registry, NotInRegistry
from gisaf.plugins import DownloadResponse, manager as plugin_manager
from gisaf.exporters import export_with_fiona, export_with_pyshp

logger = logging.getLogger(__name__)

api = APIRouter(
    tags=["download"],
    # dependencies=[Depends(get_token_header)],
    responses={404: {"description": "Not found"}},
)

@api.get('/csv/{store}/{model_id}/{value}/{resample}',
         responses={
             200: {
                 "content": {
                     "text/csv": {}
                 }
             }
         })
async def download_csv(
    store: str, model_id: int, value: str, resample: str,
    db_session: db_session,
    user: Annotated[UserRead, Depends(get_current_active_user)]
    ) -> StreamingResponse:
    try:
        store_record = registry.stores.loc[store]
        model: type[GeoModel] = store_record.model
        values_model = registry.values_for_model[model][0]
    except KeyError:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)
    ## Check if there's a custom getter
    custom_getter = getattr(values_model, f'get_{value}', None)
    if custom_getter:
        df = await custom_getter(model_id)
    else:
        item = await db_session.get(model, model_id)
        df = await values_model.get_as_dataframe(item=item, with_only_columns=[value])
    if len(df) == 0:
        raise HTTPException(status.HTTP_204_NO_CONTENT)
    if resample and resample != '0':
        value_defs = [v for v in values_model.values if v['name'] == value]
        if len(value_defs) > 0 and 'agg' in value_defs[0]:
            resampling_method = value_defs[0]['agg']
        else:
            resampling_method = 'mean'
        df = df.resample(resample).agg(resampling_method)
        if len(df) > 0:
            df.reset_index(inplace=True)
    query = select(model).where(model.id == model_id).options(
        *(joinedload(jt) for jt in model.selectinload()))
    response = await db_session.exec(query)
    item = response.one()
    filename = f'"{item.caption}-{value}-{resample}.csv"'
    filename = f'{item.__class__.__name__}-{value}-{resample}.csv'
    stream = io.StringIO()
    df.reset_index().to_csv(stream,
                            index=False,
                            date_format='%d/%m/%Y %H:%M',
                            float_format=values_model.float_format)
    response = Response(stream.getvalue(),
                                 headers={
                                    'Content-Disposition': f"attachment; filename={filename}"
                                 })
    return response


@api.get('/geodata/{stores}')
async def download_geodata(
    user: Annotated[UserRead, Depends(get_current_active_user)],
    stores: str,
    format: str = 'gpkg',
    reproject: bool = False,
) -> Response:
    ## Check permissions
    try:
        layers_registry = registry.stores.loc[stores.split(',')]
    except KeyError:
        ## Empty dataframe
        layers_registry = registry.stores[0:0]

    live_stores = [
        layer for layer in stores.split(',')
        if layer.startswith('live:')
    ]

    def _check_permission(model):
        if not hasattr(model, 'downloadable_role'):
            return True
        if user is None:
            return False
        return user.has_role(model.downloadable_role)

    stores_allowed: list[str] = [
        c[0] for c in layers_registry.model.items()
        if _check_permission(c[1])
    ] # type: ignore
    store_names = ','.join(stores_allowed + live_stores)

    if len(store_names) == 0:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    formats = {
        'gpkg': (
            export_with_fiona, {
                'extension': 'gpkg',
                'driver': 'GPKG',
                'mimetype': 'application/geopackage+vnd.sqlite3',
            }
        ),
        'dxf': (
            export_with_fiona, {
                'extension': 'dxf',
                'driver': 'DXF',
                'mimetype': 'application/x-dxf',
                'filter_columns': ['geom'],
            }
        ),
        'shapefile': (
            ## Fiona/ogr has several limitations for writing shapefiles (eg. dbf, shx not supported): use export_with_pyshp
            #export_with_fiona, {
            #    'extension': 'shp',
            #    'driver': 'ESRI Shapefile',
            #    'mimetype': 'application/octet-stream',
            #}
            export_with_pyshp, {}
        ),
    }

    if format not in formats:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND,
                            detail=f"{format} not in known formats: {', '.join(formats.keys())}")

    fn, kwargs = formats[format]
    if reproject and reproject not in ['false', '0', 'no']:
        kwargs['reproject'] = True
    body, filename, content_type = await fn(store_names=store_names, **kwargs)

    headers = {
        'Content-Disposition': f'attachment; filename="{filename}"'
    }
    return Response(
        body,
        headers=headers,
        media_type=content_type,
    )


@api.get('/plugin/{name}/{store}/{id}')
async def execute_action(
        name: str,
        store: str,
        id: int,
        db_session: db_session,
        user: Annotated[UserRead, Depends(get_current_active_user)]
    ):
    """
    Download the result of an action
    """
    ## TODO: implement permissions for actions
    #await check_permission(info.context['request'], 'action')
    try:
        store_record = registry.stores.loc[store]
        model: type[GeoModel] = store_record.model
        values_model = registry.values_for_model[model][0]
    except KeyError:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)
    item = await db_session.get(model, id)
    dls: list[Downloader] = [dl for dl in plugin_manager.downloaders_stores[store]
                                  if dl.name == name]
    if len(dls) == 0:
        raise HTTPException(status.HTTP_501_NOT_IMPLEMENTED,
                            detail=f'No downloader {name} for {store}')
    elif len(dls) > 1:
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR,
                            detail=f'Too many downloaders ({len(dls)}) {name} for {store}')
    downloader = dls[0]
    result: DownloadResponse
    try:
        result = await downloader._plugin.execute(model, item)
    except Exception as err:
        logging.exception(err)
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR,
                            detail=f'Error in action: {err.args[0]}'
        )
    return StreamingResponse(
        iter([result.content]),
        headers = {
            'Content-Disposition': f'attachment; filename="{result.file_name}"'
        },
        media_type=result.content_type,
    )