import logging
from contextlib import asynccontextmanager

from fastapi import FastAPI, responses

from gisaf.config import conf
from gisaf.registry import registry
from gisaf.redis_tools import setup_redis, setup_redis_cache, shutdown_redis
from gisaf.tiles import registry as map_tile_registry
from gisaf.live import setup_live
from gisaf.admin import manager as admin_manager
from gisaf.api.main import api
from gisaf.api.geoapi import api as geoapi
from gisaf.api.admin import api as admin_api
from gisaf.api.dashboard import api as dashboard_api
from gisaf.api.map import api as map_api
from gisaf.api.download import api as download_api
from gisaf.plugins import manager as plugin_manager

logging.basicConfig(level=conf.gisaf.debugLevel)
logger = logging.getLogger(__name__)


@asynccontextmanager
async def lifespan(app: FastAPI):
    await registry.make_registry()
    await setup_redis()
    await setup_redis_cache()
    await setup_live()
    await plugin_manager.scan_plugins()
    await admin_manager.setup_admin()
    await map_tile_registry.setup()
    yield
    await shutdown_redis()
    await map_tile_registry.shutdown()

app = FastAPI(
    debug=False,
    title=conf.gisaf.title,
    version=conf.version,
    lifespan=lifespan,
    default_response_class=responses.ORJSONResponse,
)

app.include_router(api, prefix="/api")
app.include_router(geoapi, prefix="/api/gj")
app.include_router(admin_api, prefix="/api/admin")
app.include_router(dashboard_api, prefix="/api/dashboard")
app.include_router(map_api, prefix='/api/map')
app.include_router(download_api, prefix='/api/download')