from re import A
import geopandas as gpd
from shapely import from_wkb
from json import dumps

from sqlmodel import SQLModel

from gisaf.config import conf
from gisaf.models.map_bases import MaplibreStyle
from gisaf.models.info import FeatureInfo


class BaseStore(SQLModel):
    mapbox_type: str = 'symbol'
    name: str = '<Unnamed store>'
    description: str = '<Description>'
    icon: str | None = None
    mapbox_paint: dict[str, dict] | None = None
    mapbox_layout: dict[str, dict] | None = None
    attribution: str | None = None
    symbol: str = '\ue32b'
    base_gis_type: str = 'Point'
    z_index: int = 460
    cache_enabled: bool = False
    can_get_features_as_df: bool = True
    filtered_columns_on_map: list[str] = []
    status: str = 'E'
    ## TODO: count and other Model-like interface
    count: int = -1

    @classmethod
    async def get_popup(cls, df):
        return cls.__name__ + ': ' + df.index.astype('U')

    @classmethod
    async def get_properties(cls, df):
        return {}

    ## XXX: Copied from GeoModel
    ## TODO: Create a mixin for stores/models? Set Model as a subclass of Store?
    @classmethod
    async def get_geo_df(cls, where=None, crs=None, reproject=False,
                         filter_columns=False, with_popup=False, **kwargs):
        """
        Return a Pandas dataframe of all records
        :param where: where clause for the query (eg. Model.attr=='foo')
        :param crs: coordinate system (eg. 'epsg:4326') (priority over the reproject parameter)
        :param reproject: should reproject to conf.srid_for_proj
        :return:
        """
        df = await cls.get_df(where=where, **kwargs)
        df.dropna(subset=['geom'], inplace=True)
        #df.set_index('id', inplace=True)
        df.sort_index(inplace=True)
        df_clean = df[df.geom != None]
        ## Drop coordinates
        df_clean.drop(columns=set(df_clean.columns).intersection(['ST_X_1', 'ST_Y_1', 'ST_Z_1']),
                      inplace=True)
        if not crs:
            crs = conf.crs['geojson']

        #if getattr(gpd.options, 'use_pygeos', False):
        #    geometry = from_wkb(df_clean.geom)
        #else:
        #    geometry = [wkb.loads(geom) for geom in df_clean.geom]
        ## XXX: There must be a vectorized way to do this
        geom_bin = df_clean.geom.apply(lambda row: row.desc)
        geometry = from_wkb(geom_bin)
        # if not getattr(gpd.options, 'use_pygeos', False):
        #     geometry = pygeos.to_shapely(geometry)

        gdf = gpd.GeoDataFrame(
            df_clean.drop('geom', axis=1),
            crs=crs,
            geometry=geometry
        )

        if hasattr(cls, 'simplify') and cls.simplify:
            #shapely_geom = shapely_geom.simplify(simplify_tolerance / conf.geo.simplify_geom_factor,
                                                 #preserve_topology=conf.geo.simplify_preserve_topology)
            gdf['geometry'] = gdf['geometry'].simplify(
                float(cls.simplify) / conf.geo.simplify_geom_factor,
                preserve_topology=conf.geo.simplify_preserve_topology)

        if reproject:
            gdf.to_crs(crs=conf.crs['for_proj'], inplace=True)

        ## Filter out columns
        if filter_columns:
            gdf.drop(columns=set(gdf.columns).intersection(cls.filtered_columns_on_map),
                     inplace=True)

        if with_popup:
            gdf['popup'] = await cls.get_popup(gdf)

        return gdf

    @classmethod
    async def get_df(cls, where=None,
                     with_related=None, recursive=True,
                     cast=True,
                     with_only_columns=None,
                     geom_as_ewkt=False,
                     **kwargs):
        """
        Return a Pandas dataframe of all records
        Optional arguments:
        * an SQLAlchemy where clause
        * with_related: automatically get data from related columns, following the foreign keys in the model definitions
        * cast: automatically transform various data in their best python types (eg. with date, time...)
        * with_only_columns: fetch only these columns (list of column names)
        * geom_as_ewkt: convert geometry columns to EWKB (handy for eg. using upsert_df)
        :return:
        """
        raise NotImplementedError('Subclasses of BaseStore must implement get_df()')

    @classmethod
    async def get_item_params(cls, id) -> FeatureInfo:
        raise NotImplementedError('Subclasses of BaseStore must implement get_item_params()')

    @classmethod
    async def get_maplibre_style(cls) -> MaplibreStyle:
        """
        Get the mapbox style (paint, layout, attribution...)
        """
        return MaplibreStyle(
            paint=cls.mapbox_paint,
            layout=cls.mapbox_layout,
            attribution=cls.attribution
        )