from contextlib import asynccontextmanager
from typing import Annotated, Literal, Any
from collections.abc import AsyncGenerator

from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy import create_engine
from sqlalchemy.orm import joinedload, QueryableAttribute
from sqlalchemy.sql.selectable import Select
from sqlmodel import SQLModel, select
from sqlmodel.ext.asyncio.session import AsyncSession
from fastapi import Depends
# from geoalchemy2.functions import ST_SimplifyPreserveTopology
import pandas as pd
import geopandas as gpd # type: ignore

from gisaf.config import conf

engine = create_async_engine(
    conf.db.get_sqla_url(),
    echo=conf.db.echo,
    pool_size=conf.db.pool_size,
    max_overflow=conf.db.max_overflow,
)

sync_engine = create_engine(
    conf.db.get_pg_url(),
    echo=conf.db.echo,
    pool_size=conf.db.pool_size,
    max_overflow=conf.db.max_overflow,
)

async def get_db_session() -> AsyncGenerator[AsyncSession]:
    async with AsyncSession(engine) as session:
        yield session

@asynccontextmanager
async def db_session() -> AsyncGenerator[AsyncSession]:
    async with AsyncSession(engine) as session:
        yield session

def pandas_query(session, query):
    return pd.read_sql_query(query, session.connection())

def geopandas_query(session, query: Select, model, *,
                    # simplify_tolerance: float|None=None,
                    crs=None, cast=True,
                    ):
    ## XXX: I could not get the add_columns work without creating a subquery,
    ## so moving the simplification to geopandas - see in _get_df
    # if simplify_tolerance is not None:
    #     query = query.with_only_columns(*(col for col in query.columns
    #                                      if col.name != 'geom'))
    #     new_column = model.__table__.columns['geom'].ST_SimplifyPreserveTopology(
    #                                     simplify_tolerance).label('geom')
    #     query = query.add_columns(new_column)
    return gpd.GeoDataFrame.from_postgis(query, session.connection(), crs=crs)

class BaseModel(SQLModel):
    @classmethod
    def selectinload(cls) -> list[Literal['*'] | QueryableAttribute[Any]]:
        return []

    @classmethod
    async def get_df(cls, **kwargs) -> pd.DataFrame:
        return await cls._get_df(pandas_query, **kwargs)

    @classmethod
    async def get_gdf(cls, **kwargs) -> gpd.GeoDataFrame:
        return await cls._get_df(geopandas_query, model=cls, **kwargs) # type: ignore

    @classmethod
    async def _get_df(cls, method, *,
        where=None, with_related=True, with_only_columns=[],
        simplify_tolerance: float | None=None,
        preserve_topology: bool | None=None,
        **kwargs) -> pd.DataFrame | gpd.GeoDataFrame:
        async with db_session() as session:
            if len(with_only_columns) == 0:
                query = select(cls)
            else:
                columns = set(with_only_columns)
                columns.add(*(col.name for col in cls.__table__.primary_key.columns)) # type: ignore
                query = select(*(getattr(cls, col) for col in columns))
            if where is not None:
                query = query.where(where)
            ## Get the joined tables
            if with_related:
                joined_tables = cls.selectinload()
            if with_related and len(joined_tables) > 0:
                query = query.options(*(joinedload(jt) for jt in joined_tables))
                ## Get the column details of the final query
                # ffrom = query.get_final_froms()[0]
                # for col in ffrom.columns:
                #     if col.table == cls.__table__:
                #         continue
                #     print(col, col._proxies[0])
                #     pass
            df = await session.run_sync(method, query, **kwargs)
            if method is geopandas_query and simplify_tolerance is not None:
                df['geom'] = df['geom'].simplify(
                    simplify_tolerance / conf.geo.simplify_geom_factor,
                    preserve_topology=(conf.geo.simplify_preserve_topology
                                       if preserve_topology is None
                                       else preserve_topology)
                )
            ## Chamge column names to reflect the joined tables
            ## Leave the first columns unchanged, as their names come straight
            ## from the model's fields
            joined_columns = list(df.columns[len(cls.model_fields):])
            renames: dict[str, str] = {}
            # Match colum names with the joined tables
            # This uses the fact that orders of the joined tables
            # and their columns is preserved by sqlalchemy query options (joinedlaod),
            # and pandas' read_sql
            # Important: apparently, the order defined in cls.selectinload has to match the
            # order of the relationship fields defined in the model class definition
            for joined_table in joined_tables:
                target = joined_table.property.target  # type: ignore
                for col in target.columns:
                    ## Pop the column from the colujmn list and make a new name
                    renames[joined_columns.pop(0)] = f'{target.schema}_{target.name}_{col.name}'
            df.rename(columns=renames, inplace=True)
            ## Finally, set the index of the df as the index of cls
            df.set_index([c.name for c in cls.__table__.primary_key.columns],  # type: ignore
                        inplace=True)
            return df


fastapi_db_session = Annotated[AsyncSession, Depends(get_db_session)]