import asyncio
import logging
from collections import defaultdict

from fastapi import FastAPI, WebSocket, WebSocketDisconnect

# from .config import conf
from gisaf.redis_tools import store

logger = logging.getLogger(__name__)

class LiveServer:
    def __init__(self):
        self.ws_clients = defaultdict(set)

    async def setup(self, listen_to_redis=False, with_mqtt=False):
        """
        Setup for the live server
        """
        if with_mqtt:
            logger.warning('Gisaf LiveServer does not support with_mqtt: ignoring')
        if listen_to_redis:
            self.pub = store.redis.pubsub()
            await self.pub.psubscribe('live:*:json')
            asyncio.create_task(self._listen_to_redis())

    async def _listen_to_redis(self):
        """
        Subscribe the redis sub channel to all data ("live:*:json"),
        and send the messages to websockets
        """
        async for msg in self.pub.listen():
            if msg['type'] == 'pmessage':
                await self._send_to_ws_clients(msg['channel'].decode(),
                                               msg['data'].decode())

    async def _send_to_ws_clients(self, store_name, json_data):
        """
        Send the json_data to the websoclets which have subscribed
        to that channel (store_name)
        """
        if len(self.ws_clients[store_name]) > 0:
            logger.debug(f'WS channel {store_name} got {len(json_data)} bytes to send to:'
                         f' {", ".join([str(id(ws)) for ws in self.ws_clients[store_name]])}')
            for ws in self.ws_clients[store_name]:
                if ws.client_state.name != 'CONNECTED':
                    logger.debug(f'Cannot send {store_name} for WS {id(ws)}, state: {ws.client_state.name}')
                    continue
                try:
                    await ws.send_text(json_data)
                    logger.debug(f'Sent live update for WS {id(ws)}: {len(json_data)}')
                except RuntimeError as err:
                    ## The ws is probably closed, remove it from the clients
                    logger.debug(f'Cannot send live update for {store_name}: {err}')
                    del self.ws_clients[store_name]
        else:
            pass
            #logger.debug(f'WS channel {store_name} has no clients')

    def add_subscription(self, ws, store_name):
        """
        Add the websocket subscription to the layer
        """
        channel = store.get_json_channel(store_name)
        logger.debug(f'WS {id(ws)} subscribed to {channel}')
        self.ws_clients[channel].add(ws)

    def remove_subscription(self, ws, store_name):
        """
        Remove the websocket subscription to the layer
        """
        channel = store.get_json_channel(store_name)
        if ws in self.ws_clients[channel]:
            self.ws_clients[channel].remove(ws)


async def setup_live():
    global live_server
    await live_server.setup(listen_to_redis=True)

live_server = LiveServer()
