from shapely.ops import transform  # type: ignore

from shapely.geometry import (
    Point, LineString, MultiLineString, Polygon, MultiPolygon)

from gisaf.redis_tools import store as redis_store
from gisaf.models.geo_models_base import reproject_func
from gisaf.models.info import FeatureInfo, InfoItem

async def get_live_feature_info(store: str, id: str) -> FeatureInfo:
        item = await redis_store.get_feature_info(store, id)
        geom = item.geometry
        ## Reproject to projected coordinate system
        geom_reprojected = transform(reproject_func, geom)
        geoInfoItems: list[InfoItem] = []
        if isinstance(geom, Point):
            geoInfoItems.append(InfoItem(key='longitude',
                                         value=f'{geom.x:.6f}'))
            geoInfoItems.append(InfoItem(key='latitude',
                                         value=f'{geom.y:.6f}'))
            if geom.has_z:
                geoInfoItems.append(InfoItem(key='elevation (m)',
                                             value=f'{geom.z:.6f}'))
        elif isinstance(geom, (LineString, MultiLineString)):
            bounds = geom.bounds
            geoInfoItems.append(InfoItem(key='longitude',
                                         value=f'{bounds[0]:.6f} - {bounds[2]:.6f}'))
            geoInfoItems.append(InfoItem(key='latitude',
                                         value=f'{bounds[1]:.6f} - {bounds[3]:.6f}'))
            geoInfoItems.append(InfoItem(key='length (m)',
                                         value=f'{geom_reprojected.length:.2f}'))
            ## TODO: elevation for MultiLineString
            if geom.has_z and not isinstance(geom, MultiLineString):
                elevations = [cc[2] for cc in geom.coords]
                elev_min = min(elevations)
                elev_max = max(elevations)
                if elev_min == elev_max:
                    geoInfoItems.append(InfoItem(key='elevation (m)',
                                         value=f'{elev_min:.2f}'))
                else:
                    geoInfoItems.append(InfoItem(key='elevation (m)',
                                         value=f'{elev_min:.2f} - {elev_max:.2f}'))
        elif isinstance(geom, (Polygon, MultiPolygon)):
            area = geom_reprojected.area
            bounds = geom.bounds
            geoInfoItems.append(InfoItem(key='longitude',
                                         value=f'{bounds[0]:.6f} - {bounds[2]:.6f}'))
            geoInfoItems.append(InfoItem(key='latitude',
                                         value=f'{bounds[1]:.6f} - {bounds[3]:.6f}'))
            geoInfoItems.append(InfoItem(key='area (sq. m)',
                                         value=f'{area:.1f} sq. m'))
            geoInfoItems.append(InfoItem(key='area (ha)',
                                         value=f'{area / 10000:.1f} ha'))
            geoInfoItems.append(InfoItem(key='area (acre)',
                                         value=f'{area / 4046.85643005078874:.1f} acres'))
            ## TODO: elevation for MultiPolygon
            if geom.has_z and not isinstance(geom, MultiPolygon):
                if hasattr(geom, 'exterior'):
                    coords = geom.exterior.coords
                else:
                    coords = geom.coords
                elevations = [coord[2] for coord in coords]
                elev_min = min(elevations)
                elev_max = max(elevations)
                if elev_min == elev_max:
                    geoInfoItems.append(InfoItem(key='elevation (m)',
                                                 value=f'{elev_min:.2f}'))
                else:
                    geoInfoItems.append(InfoItem(key='elevation (m)',
                                                 value=f'{elev_min:.2f} - {elev_max:.2f}'))
        return FeatureInfo(
            id=id,
            itemName=item.get('popup', f'Live: {store} #{id}'),
            geoInfoItems=geoInfoItems,
            surveyInfoItems=[InfoItem(key='Note',
                                      value='Live layers do not have survey info')],
            infoItems=[
                InfoItem(key=key, value=value)
                for key, value in item.items()
                if key not in ('geometry', 'popup')
            ]
        )
