from typing import Annotated, ClassVar
from geoalchemy2 import Geometry, WKBElement
from sqlmodel import Field, BigInteger, Relationship

from gisaf.config import conf
from gisaf.models.models_base import Model
from gisaf.models.geo_models_base import GeoPointMModel, BaseSurveyModel
from gisaf.models.project import Project
from gisaf.models.category import Category
from gisaf.models.metadata import gisaf_survey, gisaf_admin

class RawSurveyModel(BaseSurveyModel, GeoPointMModel, table=True):
    __table_args__ = gisaf_survey.table_args
    __tablename__ = 'raw_survey'
    hidden: ClassVar[bool] = True

    geom: Annotated[str, WKBElement] = Field(
        sa_type=Geometry('POINTZ', dimension=3, srid=conf.geo.raw_survey.srid))
    id: int | None = Field(default=None, primary_key=True)
    project_id: int | None = Field(foreign_key=gisaf_admin.table('project.id'))
    category: str = Field(foreign_key=gisaf_survey.table('category.name'))
    #in_menu: bool = False
    project: Project = Relationship()
    category_info: Category = Relationship()

    ## XXX: Unused - calls to get_gdf have to provide this
    ## if the CRS is not standard, maybe due to an update of shapely?
    _crs = conf.geo.raw_survey.spatial_sys_ref

    @classmethod
    def selectinload(cls):
        return [
            cls.project,
            cls.category_info
        ]

    # @classmethod
    # def dyn_join_with(cls):
    #     return {
    #         'project': Project.on(cls.project_id == Project.id),
    #         'category_info': Category.on(cls.category == Category.name),
    #     }

    #id = db.Column(db.BigInteger, primary_key=True)
    ## XXX: Can remove the rest since it's is in the GeoPointSurveyModel class?
    #geom = db.Column(Geometry('POINTZ', srid=conf.raw_survey_srid))
    #date = db.Column(db.Date)
    #orig_id = db.Column(db.String)
    #status = db.Column(db.String(1))

    def __str__(self):
        return 'Raw Survey point id {:d}'.format(self.id)

    def to_row(self):
        """
        Get a list of attributes, typically used for exporting in CSV
        :return: list of attributes
        """
        return [
            self.id,
            self.easting,
            self.northing,
            self.elevation,
            self.category,
            self.surveyor,
            self.equipment,
            self.date.isoformat(),
            self.accuracy.name,
            self.category_info.status,
            self.project.name,
            self.orig_id
        ]

    def auto_import(self, session, model=None, status=None):
        """
        Automatically feed the raw_geom get_raw_survey_model_mapping
        :return:
        """
        if model is None:
            # XXX: move as module import?
            from gisaf.registry import registry
            model = registry.get_raw_survey_model_mapping().get(self.category)

        new_point = model(
            id=self.id,
            geom=self.geom,
            date=self.date,
            project_id=self.project_id,
            equip_id=self.equip_id,
            srvyr_id=self.srvyr_id,
            accur_id=self.accur_id,
            orig_id=self.orig_id,
            status=status,
        )
        session.merge(new_point)


class OriginRawPoint(Model):
    """
    Store information of the raw survey point used in the line work
    for each line and polygon shape
    Filled when importing shapefiles
    """
    __table_args__ = gisaf_survey.table_args
    __tablename__ = 'origin_raw_point'

    id: int | None = Field(default=None, primary_key=True)
    shape_table: str = Field(index=True)
    shape_id: int = Field(index=True)
    raw_point_id: int = Field(sa_type=BigInteger())

    def __repr__(self):
        return f'<models.OriginRawPoint {self.id:d} {self.shape_table:s} ' \
               f'{self.shape_id:d} {self.raw_point_id:d}>'