import logging
import re
from collections import defaultdict
from importlib.metadata import entry_points
from itertools import chain
from datetime import datetime

# from aiohttp.web_exceptions import HTTPUnauthorized
# from aiohttp_security import check_permission

from fastapi import HTTPException, status
from sqlalchemy import or_, and_
# from geoalchemy2.shape import to_shape, from_shape
# from graphene import ObjectType, String, List, Boolean, Field, Float, InputObjectType
import pandas as pd
import shapely # type: ignore
from pydantic import BaseModel

from gisaf.config import conf
from gisaf.models.store import Store  # noqa: F401
from gisaf.models.tags import Tags as TagsModel
from gisaf.models.authentication import UserRead
from gisaf.utils import upsert_df
from gisaf.models.reconcile import StatusChange
from gisaf.models.info import (
    ActionResults,
    ActionResult,
    Downloader,
    TagAction, ActionAction,
    TagActions,
    TagsStore,
    TagsStores,
    ActionsStore,
    Action,
    ActionParam, FormFieldInput,
    TaggedLayer, TaggedFeature
)
from gisaf.registry import NotInRegistry, registry

logger = logging.getLogger('Gisaf plugin manager')

class ActionPlugin:
    """
    Base class for all actions plugins.
    """
    def __init__(self, name, stores=None, stores_by_re=None,
                 icon='build', roles=None, params=None,
                 form_fields=None):
        self.name = name
        self.stores = stores or []
        self.stores_by_re = stores_by_re or []
        self.roles = roles or []
        self.params = params or {}
        self.icon = icon
        self.form_fields = form_fields or []

    async def execute(self,
                      user: UserRead,
                      features: dict[str, list],
                      params: list,
                      form_fields: list[FormFieldInput]) -> ActionResult:
        raise NotImplementedError('Action plugins must implement execute')


class TagPlugin:
    """
    Base class for tag plugins.
    Tags can have a "domain". The tag's key is actally like: 'domain:key'.
    Keys might be reg exp.
    See Link (below) for a very basic example.
    """
    key: str
    domain: str
    stores: list[str]
    stores_by_re: list[str]
    roles: list[str] | None
    save: bool
    link: str | None
    action: str | None

    def __init__(self,
                 key: str = '',
                 domain: str = '',
                 stores: list[str] | None = None,
                 stores_by_re: list[str] | None = None,
                 roles: list[str] | None = None,
                 save: bool = True,
                 link: str | None = None,
                 action: str | None = None):
        ## self._tags: instanciated tags
        self.key = key
        self.domain = domain
        self.stores = stores or []
        self.stores_by_re = stores_by_re or []
        self.roles = roles or []
        #self.keys = {tag.key: tag(self) for tag in self.tags}
        self.save = save
        self.link = link
        self.action = action

    @property
    def full_key(self):
        if self.domain:
            return f'{self.domain}:{self.key}'
        else:
            return self.key


class DownloadResponse(BaseModel):
    file_name: str
    content_type: str
    content: str


class DownloadPlugin:
    """
    Base class for all download plugins.
    """
    def __init__(self, name, stores=None, stores_by_re=None,
                 icon='cloud_download', roles=None,
                 form_fields=None):
        self.name = name
        self.stores = stores or []
        self.stores_by_re = stores_by_re or []
        self.roles = roles or []
        self.icon = icon

    async def execute(self, model, item) -> DownloadResponse:
        raise NotImplementedError(f'Missing execute in downloader {self.name}')


class DownloadCSVPlugin(DownloadPlugin):
    async def execute(self, model, item) -> DownloadResponse:
        try:
            values_models = registry.values_for_model[model]
        except KeyError:
            raise NotInRegistry
        for value_model in values_models:
            df = await value_model.get_as_dataframe(item=item)
            csv = df.to_csv(date_format='%d/%m/%Y %H:%M', float_format=value_model.float_format)
            ## TODO: implement multiple values for a model (search for values_for_model)
            break
        return DownloadResponse(
            file_name=f'{model.__name__}-id-{item.id}.csv',
            content_type='text/csv',
            content=csv
        )


class PluginManager:
    """
    Application wide manager of the plugins.
    One instance only, handled by Gisaf's process.
    """
    def setup(self):
        for entry_point in entry_points().select(group='gisaf_extras.context'):
            try:
                context = entry_point.load()
            except ModuleNotFoundError as err:
                logger.warning(err)
                continue
            # self.app.cleanup_ctx.append(context)
            logger.info(f'Added context for {entry_point.name}')

    async def scan_plugins(self) -> None:
        """
        Scan tag and action plugins from the Python entry points.
        Get all references of the tags defined in modules ad build a registry of:
        keys, domains, action names and executors, etc.
        Runs at startup; re-runing it later has not been tested.
        """
        self.tags_plugins = {}
        self.tags_models = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        self.tags_domains = defaultdict(list)

        #self.actions_plugins = {}
        self.actions_stores: dict[str, dict[str, list[ActionAction]]] = {}
        self.executors = defaultdict(list)
        self.downloaders = defaultdict(list)
        self.downloaders_stores: dict[str, list[Downloader]] = defaultdict(list)

        registered_models = registry.geom
        registered_stores = registered_models.keys()

        for entry_point in entry_points().select(group='gisaf_extras.tags'):
            try:
                tagPlugin: TagPlugin = entry_point.load()
            except ModuleNotFoundError as err:
                logger.warning(err)
                continue
            ## Keys, domains
            self.tags_domains[tagPlugin.domain].append(tagPlugin)
            stores = tagPlugin.stores

            ## Stores to which the tags apply
            for _tag in tagPlugin.stores_by_re:
                _re = re.compile(_tag)
                for store in registered_stores:
                    if _re.match(store) and store not in stores:
                        stores.append(store)

            for store in stores:
                model = registered_models.get(store)
                if not model:
                    logger.warn(f'Tag plugin {entry_point.name}: skip model {store}'
                                    ', which is not found in registry')
                    continue

                ## Actions
                if tagPlugin.action is not None:
                    self.tags_models[model][tagPlugin.domain][tagPlugin.key].append(
                        TagAction(
                            _plugin=tagPlugin,
                            name=entry_point.name,
                            roles=tagPlugin.roles,
                            link=tagPlugin.link,
                            action=tagPlugin.action,
                            save=tagPlugin.save,
                        )
                    )
            logger.info(f'Added tags plugin {entry_point.name} for {len(stores)} stores')

        for entry_point in entry_points().select(group='gisaf_extras.actions'):
            try:
                action = entry_point.load()
            except ModuleNotFoundError as err:
                logger.warning(err)
                continue
            self.executors[action.name].append(action)
            stores = action.stores
            for _action in action.stores_by_re:
                _re = re.compile(_action)
                for store in registered_stores:
                    if _re.match(store) and store not in stores:
                        stores.append(store)
            for store in stores:
                model = registered_models.get(store)
                if not model:
                    logger.warn(f'Action plugin {entry_point.name}: skip model {store}'
                                    ', which is not found in registry')
                    continue
                if store not in self.actions_stores:
                    self.actions_stores[store] = {}
                if action.name not in self.actions_stores[store]:
                    self.actions_stores[store][action.name] = []
                self.actions_stores[store][action.name].append(
                    ActionAction(
                        roles=action.roles,
                        #plugin=plugin.__class__.__name__,
                        name=action.name,
                        params=action.params,
                        icon=action.icon,
                        formFields=action.form_fields
                    )
                )
            logger.info(f'Added action plugin {entry_point.name}')

        for entry_point in entry_points().select(group='gisaf_extras.downloaders'):
            try:
                downloader: DownloadPlugin = entry_point.load()
            except ModuleNotFoundError as err:
                logger.warning(err)
                continue
            self.downloaders[downloader.name].append(downloader)
            stores = downloader.stores
            for _downloader in downloader.stores_by_re:
                _re = re.compile(_downloader)
                for store in registered_stores:
                    if _re.match(store) and store not in stores:
                        stores.append(store)
            for store in stores:
                model = registered_models.get(store)
                if not model:
                    logger.warn(f'Downloader plugin {entry_point.name}: skip model {store}'
                                    ', which is not found in registry')
                    continue
                _store = Downloader(
                    _plugin=downloader,
                    name=downloader.name,
                    roles=downloader.roles,
                    icon=downloader.icon,
                )
                self.downloaders_stores[store].append(_store)
            logger.info(f'Added downloader plugin {entry_point.name}')

        self.tagsStores = TagsStores(
            stores=[
                TagsStore(
                    store=model.get_store_name(),
                    tagActions=[
                        TagActions(domain=domain, key=key, actions=actions)
                        for domain, tag_keys in tag_domains.items()
                        for key, actions in tag_keys.items()
                    ]
                )
                for model, tag_domains in self.tags_models.items()
            ]
        )

        self.actionsStores = [
            ActionsStore(
                store=store,
                actions=[
                    Action(
                        name=name,
                        roles=[rr for rr in set(
                            [r for r in chain.from_iterable(
                                [aa.roles for aa in action_actions]
                            )]
                        )],
                        params=[rr for rr in set(
                            [r for r in chain.from_iterable(
                                [aa.params for aa in action_actions]
                            )]
                        )],
                    )
                    for name, action_actions in actions.items()
                ]
            )
            for store, actions in self.actions_stores.items()
        ]

    async def do_tag_action(self, request, store, id, plugin_name, value, action):
        logger.warning('FIXME: in do_tag_action: self.tags_plugins is never populated!')
        ## FIXME: self.tags_plugins is never populated!
        plugin = self.tags_plugins[plugin_name]
        ## Check permission: roles
        ## XXX: no way to check several roles?
        if len(plugin.tag.roles) > 0:
            allowed = False
            for role in plugin.tag.roles:
                try:
                    await check_permission(request, role)
                except HTTPUnauthorized as err:
                    pass
                else:
                    allowed = True
            if not allowed:
                raise Exception('Not authorized')
        ## Execute
        result = await plugin.tag.execute(request, store, int(id), value)
        return result

    async def execute_tag_mutation(self, features, key, value):
        """
        Execute tag action on the features (store and ids)
        """
        results = []
        if ':' in key:
            domain, key = key.split(':', 1)
        else:
            domain = ''
        try:
            plugins = self.tags_domains[domain]
        except KeyError:
            logger.warn(f"No tag domain '{domain}' defined by any plugin")
            return
        for plugin in plugins:
            if key != plugin.key:
                continue
            features_for_action = {store: ids
                                   for store, ids in features.items()
                                   if store in plugin.stores}
            if features_for_action:
                if plugin.save:
                    await self.save_tags(features_for_action, domain, key, value)
                #results.append(await action(features_for_action, key, value))
        return ', '.join(results)

    async def execute_action(self,
                             user: UserRead,
                             features: dict,
                             name: str,
                             params: list[ActionParam | None],
                             form_fields: list[FormFieldInput]
        ) -> ActionResults:
        """
        Execute the plugin action by calling the executor's execute function.
        It is up to the plugin action to check for security, using eg:

        from aiohttp_security import check_permission
        ...
        await check_permission(request, 'role')
        """
        results = ActionResults(name=name)
        for executor in self.executors[name]:
            ## TODO: get features from DB?
            ## Check permission
            if executor.roles:
                if user is None or not any([user.has_role(role) for role in executor.roles]):
                    raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
            result: ActionResult
            result = await executor.execute(user, features, params, form_fields)
            results.actionResults.append(result)
        return results

        #for store, ids in all_features.items():
        #    actions = self.actions_stores[store]
        #    for action in actions[name]:
        #        action._action.execute(ids, params)
        #        pass
        #for plugin in plugins:
        #    for action in action_plugin.actions:
        #        for executor in action_plugin.actions:
        #            result = executor.execute(features, params)
        #            pass

    async def save_tags(self, features, domain, key, value):
        """
        Standard tagging - should be called explicitely by plugin actions
        """
        filter = [
            and_(TagsModel.store == store, TagsModel.ref_id.in_(ids))
            for store, ids in features.items()
        ]
        tags_df = await TagsModel.get_df(where=or_(*filter), geom_as_ewkt=True)
        tags_df['tags'] = tags_df.tags.astype('object')

        for store, ids in features.items():
            model = self.app['registry'].stores.loc[store, 'model']
            feature_df = await model.get_df(where=model.id.in_(ids))
            feature_df['centroid'] = shapely.centroid(shapely.from_wkb(feature_df['geom']))
            tagged_feature_df = tags_df[tags_df.store==store].merge(
                feature_df[['id', 'centroid']],
                left_on='ref_id', right_on='id',
                suffixes=('', '_feature'),
                how='outer', indicator=True,
            )

            if domain:
                full_key = ':'.join([domain, key])
            else:
                full_key = key
            ## Fill missing values (no existing tags)
            ## Insert without pk (id) for new tags records
            new_tagged_feature_df = tagged_feature_df[tagged_feature_df['_merge']=='right_only'].drop(columns='id')
            new_tagged_feature_df['tags'] = [{full_key: value}] * len(new_tagged_feature_df)
            new_tagged_feature_df['ref_id'] = new_tagged_feature_df['id_feature']
            new_tagged_feature_df['store'] = store
            new_tagged_feature_df['geom'] = shapely.to_wkb(new_tagged_feature_df['centroid'], hex=True, include_srid=True)

            ## Existing tags: add key: value
            existing_tagged_feature_df = tagged_feature_df[tagged_feature_df['_merge']!='right_only']
            for row in existing_tagged_feature_df.itertuples():
                row.tags[full_key] = value

            ## Save
            await upsert_df(new_tagged_feature_df, TagsModel)
            await upsert_df(existing_tagged_feature_df, TagsModel)



"""Generic tag plugin for creating a link with a simple URL"""
link_tag = TagPlugin(
    key='link',
    stores_by_re=[''],  ## Match everything
    link='{value}',
)


class ChangeFeatureStatus(ActionPlugin):
    async def execute(self, request, all_features, params, **kwargs):
        results = ActionResults(actionResults=[])
        registry = request.app['registry']
        status = kwargs['Status']['value']
        for store, ids in all_features.items():
            model = registry.stores.loc[store, 'model']
            raw_model = registry.stores.loc[store, 'raw_survey_model']

            features = await model.get_df(where=model.id.in_(ids), geom_as_ewkt=True)
            raw_features = await raw_model.get_df(where=raw_model.id.in_(ids), geom_as_ewkt=True)

            ## Find the features with a status to actually change
            features = features.loc[features.status!=status]
            features['original_status'] = features['status']
            raw_features = raw_features.loc[raw_features.status!=status]

            features['status'] = status
            raw_features['status'] = status
            await upsert_df(features, model)
            await upsert_df(raw_features, raw_model)

            ## Store the status change record
            for row in features.itertuples(index=False):
                await StatusChange.create(
                    store=store,
                    ref_id=row.id,
                    original=row.original_status,
                    new=status,
                    time=datetime.now()
                )
            results.actionResults.append(ActionResult(message=f'Changed status of {len(features)} {store}(s)'))

        return results


async def create_tags(features, keys, values):
    """
    Create the tags (keys: values) for the features.
    Return a list of dataframes
    """
    from gisaf.registry import registry
    result = []
    for store, ids in features.items():
        model = registry.stores.loc[store, 'model']
        tags_df = await TagsModel.get_df(
            where=and_(TagsModel.store == store, TagsModel.ref_id.in_(ids)),
            geom_as_ewkt=True
        )
        feature_df = await model.get_df(where=model.id.in_(ids))
        feature_df['centroid'] = shapely.centroid(shapely.from_wkb(feature_df['geom']))
        tagged_feature_df = tags_df[tags_df.store==store].merge(
            feature_df[['id', 'centroid']],
            left_on='ref_id', right_on='id',
            suffixes=('', '_feature'),
            how='outer', indicator=True,
        )

        ## Fill missing values (no existing tags)
        ## Insert without pk (id) for new tags records
        key_values = dict(zip(keys, values))
        new_tagged_feature_df = tagged_feature_df[tagged_feature_df['_merge']=='right_only'].drop(columns='id')
        new_tagged_feature_df['tags'] = [key_values] * len(new_tagged_feature_df)
        new_tagged_feature_df['ref_id'] = new_tagged_feature_df['id_feature']
        new_tagged_feature_df['store'] = store
        new_tagged_feature_df['geom'] = shapely.to_wkb(new_tagged_feature_df['centroid'], hex=True, include_srid=True)

        ## Existing tags: add key: value
        existing_tagged_feature_df = tagged_feature_df[tagged_feature_df['_merge']!='right_only']
        for row in existing_tagged_feature_df.itertuples():
            for key, value in key_values.items():
                row.tags[key] = value

        ## Save
        new_tags_df = await upsert_df(new_tagged_feature_df, TagsModel)
        await upsert_df(existing_tagged_feature_df, TagsModel)
        result.append(pd.concat([existing_tagged_feature_df, new_tags_df]))

    return pd.concat(result)

from gisaf.utils import ToMigrate
logger.warning(ToMigrate('plugins.change_feature_status (graphql)'))
change_feature_status = ChangeFeatureStatus(
    name='Change status',
    stores_by_re=[f"{conf.survey.db_schema}"],
    roles=['Change status'],
    form_fields=[
        # FormField(
        #     name='Status',
        #     type='string'
        # )
    ]
)


manager = PluginManager()
