"""
The web API for Gisaf scheduler
"""
import logging
import asyncio
from datetime import datetime
from uuid import UUID
from typing import List

from fastapi import Request, WebSocket
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
from redis import asyncio as aioredis
from pandas import DataFrame

from gisaf.live import live_server
from gisaf.scheduler import GSFastAPI


logger = logging.getLogger(__name__)

app = GSFastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


class Subscriber:
    ## See https://gist.github.com/appeltel/fd3ddeeed6c330c7208502462639d2c9
    def __init__(self, hub):
        self.hub = hub
        self.queue = asyncio.Queue()

    def __enter__(self):
        self.hub.subscribers.add(self.queue)
        return self.queue

    def __exit__(self, type, value, traceback):
        self.hub.subscribers.remove(self.queue)


class JobEvent(BaseModel):
    jobId: str | UUID
    time: datetime | None
    status: str
    msg: str
    nextRunTime: datetime | None


class Feature(BaseModel):
    store: str
    ## XXX: Using "id" gives very strange issue with apollo client
    id_: str


class Job_(BaseModel):
    id: str
    name: str
    type: str
    schedParams: str
    nextRunTime: datetime | None
    lastRun: JobEvent | None
    features: List[Feature]


class Task_(BaseModel):
    id: str
    name: str
    type: str
    schedParams: str
    nextRunTime: datetime | None
    lastRunTime: datetime | None
    features: list[Feature]


def df_as_ObjectTypes(df):
    """
    Utility function that returns List(Feature) graphql from a dataframe.
    The dataframe must contain a 'store' column and the feature ids as index.
    """
    if not isinstance(df, DataFrame):
        return []
    if 'store' not in df.columns:
        # logger.warning(f'no store in get_feature_ids() for job "{job.name}"')
        return []
    return [
        Feature(id_=str(f[0]), store=f.store)
        for f in df.itertuples(index=True)
    ]


@app.websocket('/events')
async def scheduler_ws(
    ws: WebSocket,
    ):
    """
    Websocket for scheduler updates
    """
    #session = await get_session(request)
    #js = request.app.js
    await ws.accept()
    while True:
        # msg_text = await ws.receive_text()
        msg_data = await ws.receive_json()
        #await websocket.send_text(f"Message text was: {data}")
        if 'message' in msg_data:
            if msg_data['message'] == 'subscribe':
                live_server.add_subscription(ws, 'admin:scheduler')
                ws.app.js.add_subscription(ws)
            elif msg_data['message'] == 'unsubscribe':
                live_server.remove_subscription(ws, 'admin:scheduler')
                ws.app.js.delete_subscription(ws)


@app.websocket('/subscriptions')
async def subscriptions(ws: WebSocket):
    await ws.accept()
    while True:
        msg_data = await ws.receive_json()
        if msg.type == WSMsgType.TEXT:
            if msg.data == 'close':
                await ws.close()
            else:
                await ws.send_str(msg.data + '/answer')
        elif msg.type == WSMsgType.ERROR:
            print('ws connection closed with exception %s' % ws.exception())


@app.get('/time')
async def get_time():
    return datetime.now()


@app.get('/jobs')
async def get_jobs(request: Request) -> list[Task_]:
    app: GSFastAPI = request.app
    tasks = {task.id: task for task in await app.js.scheduler.data_store.get_tasks()}
    tasks_ = []
    for schedule in await app.js.scheduler.data_store.get_schedules():
        task = tasks[schedule.task_id]
        task_ = app.js.tasks[schedule.id]
        tasks_.append(
            Task_(
                id=task.id,
                name=task.id,
                type=schedule.trigger.__class__.__name__,
                schedParams='',
                lastRunTime=schedule.last_fire_time,
                nextRunTime=schedule.next_fire_time,
                features=df_as_ObjectTypes(task_.features),
            )
        )
    return tasks_


# async def setup_app_session(app):
#     """
#     Setup a redis pool for session management
#     Not related to the redis connection used by Gisaf
#     """
#     redis = aioredis.from_url('redis://localhost')
#     redis_storage = RedisStorage(redis)
#     session_identity_policy = SessionIdentityPolicy()
#     setup_session(app, redis_storage)