"""
mbtile server

Instructions (example):

cd map  ## Matches tilesBaseDir in config

curl -O http://download.geofabrik.de/asia/india/southern-zone-latest.osm.pbf

tilemaker \
    --config /usr/local/lib/gisaf_src/tilemaker_src/tilemaker/resources/config-openmaptiles.json \
    --process /usr/local/lib/gisaf_src/tilemaker_src/tilemaker/resources/process-openmaptiles.lua \
    --input southern-zone-latest.osm.pbf \
    --output southern-zone-latest.mbtile


----

Get the style from https://github.com/openmaptiles, eg.
curl -o osm-bright-full.json https://raw.githubusercontent.com/openmaptiles/osm-bright-gl-style/master/style.json
## Minify json:
python -c 'import json, sys;json.dump(json.load(sys.stdin), sys.stdout)'  < osm-bright-full.json > osm-bright.json

And copy the style to the gisaf_map.map_base_style table,
with the name matching the file name
(TODO: this would need de-coupling the source mbtile and the style)

----

Get the sprites from openmaptiles:`

cd tiles  ## Matches tilesSpriteBaseDir in config

curl -O 'https://openmaptiles.github.io/osm-bright-gl-style/sprite.png'
curl -O 'https://openmaptiles.github.io/osm-bright-gl-style/sprite.json'
curl -O 'https://openmaptiles.github.io/osm-bright-gl-style/sprite@2x.png'
curl -O 'https://openmaptiles.github.io/osm-bright-gl-style/sprite@2x.json'

TODO: TO migrate - 3/1/2024: this was copied from legacy gisaf without change
See: https://github.com/gis-ops/tutorials/blob/master/webservices/fastapi/fastapi_auth_vector_tiles.md
"""

from pathlib import Path
from json import loads, dumps
import logging

from fastapi import FastAPI, Response, responses, HTTPException, status
from fastapi.staticfiles import StaticFiles
import aiosqlite

from gisaf.config import conf

logger = logging.getLogger('gisaf tile server')

api = FastAPI(
    default_response_class=responses.ORJSONResponse,
)

OSM_ATTRIBUTION = '<a href=\"http://www.openstreetmap.org/about/\" target=\"_blank\">&copy; OpenStreetMap contributors</a>'

class MBTiles:
    def __init__(self, file_path, style_name):
        self.file_path = file_path
        self.name = style_name
        self.scheme = 'tms'
        self.etag = f'W/"{hex(int(file_path.stat().st_mtime))[2:]}"'

    async def connect(self):
        self.db = await aiosqlite.connect(self.file_path)
        self.metadata = {}
        try:
            async with self.db.execute('select name, value from metadata') as cursor:
                async for row in cursor:
                    self.metadata[row[0]] = row[1]
        except aiosqlite.DatabaseError as err:
            logger.warning(f'Cannot read {self.file_path}, will not be able to serve tiles (error: {err.args[0]})')

        self.metadata['bounds'] = [float(v) for v in self.metadata['bounds'].split(',')]
        self.metadata['maxzoom'] = int(self.metadata['maxzoom'])
        self.metadata['minzoom'] = int(self.metadata['minzoom'])

    async def get_style(self, style_record, request):
        """
        Generate on the fly the style
        """
        if conf.map.tileServer.useRequestUrl:
            base_url = request.url.parent
        else:
            base_url = conf.map.tileServer.spriteBaseUrl
        base_tiles_url = f"{base_url}/tiles/{self.name}"
        scheme = self.scheme
        ## TODO: avoid parse and serialize at every request
        layers = loads(style_record['style'])['layers']
        for layer in layers:
            if 'source' in layer:
                layer['source'] = 'gisafTiles'
        resp = {
            'basename': self.file_path.stem,
            #'center': self.center,
            'description': f'Extract of {self.file_path.stem} from OSM, powered by Gisaf',
            'format': self.metadata['format'],
            'id': f'gisaftiles_{self.name}',
            'maskLevel': 5,
            'name': self.name,
            #'pixel_scale': 256,
            #'planettime': '1499040000000',
            'tilejson': '2.0.0',
            'version': 8,
            'glyphs': f"/assets/fonts/glyphs/{{fontstack}}/{{range}}.pbf",
            'sprite': f"{base_url}{conf.map.tileServer.spriteUrl}",
            'sources': {
                'gisafTiles': {
                    'type': 'vector',
                    'tiles': [
                        f'{base_tiles_url}/{{z}}/{{x}}/{{y}}.pbf',
                    ],
                    'maxzoom': self.metadata['maxzoom'],
                    'minzoom': self.metadata['minzoom'],
                    'bounds': self.metadata['bounds'],
                    'scheme': scheme,
                    'attribution': OSM_ATTRIBUTION,
                    'version': self.metadata['version'],
                }
            },
            'layers': layers,
        }
        return resp


class MBTilesRegistry:
    mbtiles: dict[str, MBTiles]
    async def setup(self):
        """
        Read all mbtiles, construct styles
        """
        self.mbtiles = {}
        for file_path in Path(conf.map.tileServer.baseDir).glob('*.mbtiles'):
            mbtiles = MBTiles(file_path, file_path.stem)
            self.mbtiles[file_path.stem] = mbtiles
            await mbtiles.connect()

    async def shutdown(self):
        """
        Tear down the connection to the mbtiles files
        """
        for mbtiles in self.mbtiles.values():
            await mbtiles.db.close()


@api.get('/{style_name}/{z}/{x}/{y}.pbf')
async def get_tile(request, style_name: str, z:int, x: int, y: int,
                   response: Response):
    """
    Return the specific tile
    """
    if style_name not in registry.mbtiles:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)
    mbtiles = registry.mbtiles[style_name]

    if request.headers.get('If-None-Match') == mbtiles.etag:
        request.not_modified = True
        return {}

    response.headers['Content-Encoding'] = 'gzip'
    response.headers['Content-Type'] = 'application/octet-stream'
    request.response_etag = mbtiles.etag
    async with mbtiles.db.execute('select tile_data from tiles where zoom_level=? and tile_column=? and tile_row=?',
                           (z, x, y)) as cursor:
        async for row in cursor:
            return row[0]
    raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)


#@routes.get('/sprite/{name:\S+}')
#async def get_sprite(request):


@api.get('/{style_name}')
async def get_style(request, style_name: str):
    """
    Return the base style.
    Note that normal operations are processed through graphql (resolve_base_style)
    :param request:
    :return:
    """
    if style_name not in registry.mbtiles:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)
    mbtiles = registry.mbtiles[style_name]
    return await mbtiles.get_style()


registry = MBTilesRegistry()

api.mount("/sprite",
          StaticFiles(directory=conf.map.tileServer.spriteBaseDir),
          name="sprites")
